# -*- coding: UTF-8 -*-
# @Time    : 2021/11/13 6:13 下午
# @File    : ensemble_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import os
import shutil
import signal
from antgo.crowdsource.base_server import *
from antgo.utils import logger
import sys
import uuid
import json
import time
import math
import numpy as np
from io import BytesIO
from tornado.concurrent import run_on_executor
from concurrent.futures import ThreadPoolExecutor
import threading
import msgpack
import msgpack_numpy as ms


def softmax(x, axis=None):
  x = x - x.max(axis=axis, keepdims=True)
  y = np.exp(x)
  return y / y.sum(axis=axis, keepdims=True)


MB = 1024 * 1024
GB = 1024 * MB
MAX_STREAMED_SIZE = 1*GB
@tornado.web.stream_request_body
class AverageApiHandler(BaseHandler):
  executor = ThreadPoolExecutor(16)

  def __init__(self, *args, **kwargs):
    super(AverageApiHandler, self).__init__(*args, **kwargs)
    self.save_seek = 0
    self.worker_prefix = None
    self.worker_number = None
    self.file_folder = None
    self.file_name = None
    self.file_id = None

  @property
  def static_folder(self):
    return self.settings.get('static_path')

  def prepare(self):
    # 设置最大流式数据量
    self.request.connection.set_max_body_size(MAX_STREAMED_SIZE)

    # 根据ID，设置数据临时文件
    headers = self.request.headers
    self.worker_prefix = headers['worker_prefix']
    self.worker_number = headers['worker_number']
    self.file_id = headers['file_id']
    self.file_folder = os.path.join(self.static_folder, f"{self.worker_prefix}-{self.worker_number}")
    if not os.path.exists(self.file_folder):
      os.makedirs(self.file_folder)
    self.file_name = f"{headers['id']}.{headers['name']}"    # variable name.id

  def data_received(self, chunk):
      if self.file_name:
        if not os.path.exists(os.path.join(self.file_folder, self.file_name)):
          with open(os.path.join(self.file_folder, self.file_name), 'wb') as up:
              pass

        with open(os.path.join(self.file_folder, self.file_name), 'rb+') as up:
          up.seek(self.save_seek)
          up.write(chunk)
          self.save_seek = self.save_seek + len(chunk)

  def cal_shannon_entropy(self, preds, axis):
    uncertainty = -1.0 * np.sum(preds * np.log(preds + 1e-6), axis=axis, keepdims=True)
    return uncertainty

  @run_on_executor
  def avg(self, sample_id, name, value, weight):
    self.cond.acquire()
    # 转换到numpy数据
    if 'sample' not in self.db:
      self.db['sample'] = {}

    if sample_id not in self.db['sample']:
      self.db['sample'][sample_id] = {}

    if name not in self.db['sample'][sample_id]:
      self.db['sample'][sample_id][name] = {'data': [], 'count': self.worker_num, 'weight': []}

    self.db['sample'][sample_id][name]['data'].append(value)
    self.db['sample'][sample_id][name]['weight'].append(weight)

    while True:
      if len(self.db['sample'][sample_id][name]['data']) == self.worker_num:
        self.cond.notify()
        break

      if len(self.db['sample'][sample_id][name]['data']) != self.worker_num:
        self.cond.wait()

    self.cond.release()

    data = None
    if self.uncertain_vote_cfg is not None and \
        len(self.db['sample'][sample_id][name]['data']) > 1 and \
        self.db['sample'][sample_id][name]['data'][0].shape[self.uncertain_vote_cfg['axis']] > 1:
      # 计算不确定度，并加权平均
      uncertain_axis = self.uncertain_vote_cfg['axis']
      all_uncertainty_org = []
      all_uncertainty_region = []
      for sample_data, sample_weight in \
          zip(self.db['sample'][sample_id][name]['data'], self.db['sample'][sample_id][name]['weight']):
        # uncertainty_org 1,1,H,W
        uncertainty_org = self.cal_shannon_entropy(sample_data, uncertain_axis)
        uncertainty_region = np.ones(uncertainty_org.shape, dtype=np.float32)
        if self.uncertain_vote_cfg['thres'] != -1:
          uncertainty_region = (uncertainty_org > self.uncertain_vote_cfg['thres']).astype(np.float32)

        all_uncertainty_org.append(uncertainty_org)
        all_uncertainty_region.append(uncertainty_region)

      all_uncertainty_org = np.concatenate(all_uncertainty_org, axis=uncertain_axis)
      all_uncertainty_org = softmax(all_uncertainty_org, uncertain_axis)
      # certainty prob
      all_certainty_org = \
        np.split(1.0 - all_uncertainty_org,
                 indices_or_sections=all_uncertainty_org.shape[uncertain_axis],
                 axis=uncertain_axis)

      reweight_sample_weight = []
      for sample_weight, sample_certainty, sample_limited_region in \
          zip(self.db['sample'][sample_id][name]['weight'], all_certainty_org, all_uncertainty_region):
        certainty = sample_certainty * sample_limited_region + (1.0-sample_limited_region)
        reweight_sample_weight.append(sample_weight * certainty)

      weighted_data = 0.0
      weighted_norm = 0.0
      for sample_data, sample_weight in zip(self.db['sample'][sample_id][name]['data'],
                                            reweight_sample_weight):
        weighted_data += sample_data * sample_weight
        weighted_norm += sample_weight

      data = weighted_data / weighted_norm
    else:
      # 加权平均
      weighted_data = 0.0
      weighted_norm = 0.0
      for sample_data, sample_weight in zip(self.db['sample'][sample_id][name]['data'],
                                            self.db['sample'][sample_id][name]['weight']):
        weighted_data += sample_data * sample_weight
        weighted_norm += sample_weight

      data = weighted_data / weighted_norm

    # 清空缓存
    self.lock.acquire()
    self.db['sample'][sample_id][name]['count'] -= 1
    if self.db['sample'][sample_id][name]['count'] == 0:
      self.db['sample'][sample_id][name] = {'data': [], 'count': self.worker_num}

    self.lock.release()
    return data

  @gen.coroutine
  def post(self, *args, **kwargs):
    # id,name,value
    id = self.request.headers.get('id')
    name = self.request.headers.get('name')
    weight = (float)(self.request.headers.get('weight'))
    feedback = self.request.headers.get('feedback')

    if not os.path.exists(os.path.join(self.file_folder, self.file_name)):
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # 加载数据
    with open(os.path.join(self.file_folder, self.file_name), 'rb') as fp:
      value = msgpack.unpackb(fp.read(), object_hook=ms.decode)

    if not self.data_record:
      # 删除临时文件
      os.remove(os.path.join(self.file_folder, self.file_name))

    # 计算加权值
    ave_result = yield self.avg(id, name, value, weight)

    # 下载数据服务
    if feedback == 'True':
      ave_result = msgpack.packb(ave_result, default=ms.encode)
      self.set_header('Content-Type', 'application/octet-stream')
      self.set_header('Content-Disposition', f'attachment; filename={self.file_id}')
      with BytesIO(ave_result) as f:
        while True:
          data = f.read(MB)
          if not data:
              break
          self.write(data)

    #
    self.finish()

  @property
  def worker_num(self):
    return self.settings['worker_num']

  @property
  def cond(self):
    return self.settings['cond']

  @property
  def lock(self):
    return self.settings['lock']

  @property
  def uncertain_vote_cfg(self):
    return self.settings.get('uncertain_vote_cfg', None)

  @property
  def data_record(self):
    return self.settings.get('data_record', False)


@tornado.web.stream_request_body
class DataApiHandler(BaseHandler):
  @property
  def dump_dir(self):
    return self.settings.get('dump_dir', None)

  @gen.coroutine
  def get(self, file_name):
    # 提供变量数据服务
    #   ave_result = msgpack.packb(ave_result, default=ms.encode)
    file_name = file_name[1:]
    file_path = os.path.join(self.dump_dir, 'ensemble', 'merge', file_name)
    if not os.path.exists(file_path):
      self.set_status(404)
      return

    self.set_header('Content-Type', 'application/octet-stream')
    self.set_header('Content-Disposition', f'attachment; filename={file_name}')
    with open(file_path, 'rb') as f:
      while True:
        data = f.read(MB)
        if not data:
            break
        self.write(data)

    self.finish()


class LiveApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def ensemble_server_start(dump_dir, server_port, worker_num, uncertain_vote_cfg=None, enable_data_record=False):
  # uncertain_vote_cfg {'axis': 1, 'thres': 0}
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)

  # 0.step define http server port
  define('port', default=server_port, help='run on port')

  try:
    # 静态资源目录
    static_dir = os.path.join(dump_dir, 'ensemble')
    if not os.path.exists(os.path.join(static_dir, 'static')):
      os.makedirs(os.path.join(static_dir, 'static'))

    # 2.step launch web server
    db = {}
    settings = {
      'static_path': os.path.join(static_dir, 'static'),
      'dump_dir': dump_dir,
      'port': server_port,
      'cookie_secret': str(uuid.uuid4()),
      'worker_num': worker_num,
      'cond': threading.Condition(),
      'lock': threading.Lock(),
      'db': db,
      'uncertain_vote_cfg': uncertain_vote_cfg,
      'data_record': enable_data_record
    }

    app = tornado.web.Application(handlers=[(r"/ensemble-api/avg/", AverageApiHandler),
                                            (r"/ensemble-api/live/", LiveApiHandler),
                                            (r"/ensemble-api/data/(.*)", DataApiHandler)],
                                  **settings)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)

    logger.info('ensemble is providing server on port %d' % server_port)
    tornado.ioloop.IOLoop.instance().start()
    logger.info('ensemble stop server')
  except GracefulExitException:
    logger.info('ensemble server exit')
    sys.exit(0)
  except KeyboardInterrupt:
    pass


# -*- coding: UTF-8 -*-
# @Time    : 2021/1/8 7:23 下午
# @File    : interactcontext.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.context import *
from antgo.ant.train import *
from antgo.ant.challenge import *
from antgo.ant.demo import *
from antgo.ant.browser import *
from antgo.ant.batch import *
from antgo.ant.ensemble import *


class InteractContext(Context):
  def __init__(self):
    super(InteractContext, self).__init__(interact_mode=True)

  def __prepare_context(self, name, param, **kwargs):
    # 1.step 获取基本配置信息
    config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
    if not os.path.exists(config_xml):
      # 创建默认配置
      if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo')):
        os.makedirs(os.path.join(os.environ['HOME'], '.config', 'antgo'))

      env = Environment(loader=FileSystemLoader('/'.join(os.path.realpath(__file__).split('/')[0:-1])))
      config_template = env.get_template('config.xml')
      config_data = {
        'DATASET_FACTORY': os.path.join(os.environ['HOME'], '.config', 'antgo', 'factory'),
        'USER_TOKEN': ''
      }
      config_content = config_template.render(**config_data)
      with open(os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml'), 'w') as fp:
        fp.write(config_content)

      if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo', 'factory')):
        os.makedirs(os.path.join(os.environ['HOME'], '.config', 'antgo', 'factory'))

      logger.warn('Build default antgo config in ~/.config/antgo/config.xml')

    Config.parse_xml(config_xml)
    factory = getattr(Config, 'factory', None)
    assert factory is not None

    if not os.path.exists(Config.data_factory):
      os.makedirs(Config.data_factory)
    if not os.path.exists(Config.task_factory):
      os.makedirs(Config.task_factory)

    self.name = name
    self.data_factory = Config.data_factory
    self.task_factory = Config.task_factory
    self.api_token = kwargs.get('token', Config.server_user_token)
    logger.info(f'Use experiment data_factory {self.data_factory}')
    logger.info(f'Use experiment task_factory {self.task_factory}')
    logger.info(f'Use experiment api token {self.api_token}')

    self.main_folder = os.path.abspath(os.curdir)
    if self.data_factory == os.path.join(os.environ['HOME'], '.config', 'antgo', 'factory'):
      # 当前正在使用默认配置
      logger.warn('Current use default factory, please shell "antgo config --dataset=xxx --token=xxx"')

    self.params = {}
    if type(param) == str:
      param = yaml.load(open(param, 'r'))
      self.params = param
    else:
      self.params = param

    if 'system' not in self.params._params:
      self.params._params.update({'system': {}})
    if 'research' not in self.params._params['system']:
      self.params._params['system']['research'] = ''
    if 'skip_training' not in self.params._params['system']:
      self.params._params['system']['skip_training'] = False
    if 'ip' not in self.params._params['system']:
      self.params._params['system']['ip'] = '127.0.0.1'
    if 'port' not in self.params._params['system']:
      self.params._params['system']['port'] = 8901
    if 'devices' not in self.params._params['system']:
      self.params._params['system']['devices'] = ''
    if 'running_platform' not in self.params._params['system']:
      self.params._params['system']['running_platform'] = ''
    if 'unlabel' not in self.params._params['system']:
      self.params._params['system']['unlabel'] = False
    if 'candidate' not in self.params._params['system']:
      self.params._params['system']['candidate'] = False
    if 'ext_params' not in self.params._params['system']:
      self.params._params['system']['ext_params'] = {}

    # 2.step 设置来自实验名称
    self.dump_dir = os.path.join(os.path.abspath(os.curdir), 'dump')
    if not os.path.exists(self.dump_dir):
      os.makedirs(self.dump_dir)

    from_experiment = kwargs.get('from_experiment', None)
    if from_experiment is not None:
      experiment_path = os.path.join(self.dump_dir, from_experiment)
      if not os.path.exists(experiment_path):
        logger.error('Couldnt find experiment %s' % from_experiment)
        raise IOError

      if os.path.exists(os.path.join(experiment_path, 'train')):
        self.from_experiment = os.path.join(experiment_path, 'train')
      elif os.path.exists(os.path.join(experiment_path, 'inference')):
        self.from_experiment = os.path.join(experiment_path, 'inference')
      else:
        self.from_experiment = experiment_path

    # 3.step 设置默认任务
    task = kwargs.get('task', None)
    if task is not None:
      if os.path.exists(os.path.join(self.task_factory, task)):
        task = os.path.join(self.task_factory, task)
      else:
        task = os.path.join(self.main_folder, task)

    dataset = kwargs.get('dataset', None)
    if task is None and dataset is not None:
      # build default task
      with open(os.path.join(self.task_factory, '%s.xml' % name), 'w') as fp:
        task_content = '<task><task_name>%s</task_name><task_type>%s</task_type><task_badcase><badcase_num>%d</badcase_num><badcase_category>%d</badcase_category></task_badcase>' \
                       '<input>' \
                       '<source_data><data_set_name>%s</data_set_name>' \
                       '</source_data>' \
                       '<estimation_procedure><type>%s</type></estimation_procedure>' \
                       '<evaluation_measures><evaluation_measure>%s</evaluation_measure></evaluation_measures>' \
                       '</input>' \
                       '</task>' % (
                       name, '', 0, 0,dataset, '', '')
        fp.write(task_content)
      task = os.path.join(self.task_factory, '%s.xml' % name)

    self.dataset = dataset
    self.task = task

  @contextmanager
  def Train(self, exp_name, exp_param, **kwargs):
    # 1.step 配置基本信息
    self.__prepare_context(exp_name, exp_param, **kwargs)
    # 2.step 创建时间戳
    time_stamp = timestamp()
    try:
      # 6.step 创建训练过程调度对象
      train = AntTrain(self,
                       exp_name,
                       self.data_factory,
                       self.dump_dir,
                       self.api_token,
                       self.task,
                       self.dataset,
                       time_stamp=time_stamp,
                       main_folder=self.main_folder)
      # 训练控制
      train.start()
      yield train

      # 日志更新
      mlogger.update()

      # 环境回收
      train.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Challenge(self, exp_name, exp_param, benchmark='', **kwargs):
    # 1.step 配置基本信息
    self.__prepare_context(exp_name, exp_param, **kwargs)

    try:
      # 5.step 创建挑战过程调度对象
      challenge = AntChallenge(self,
                               exp_name,
                               self.data_factory,
                               self.dump_dir,
                               self.api_token,
                               self.task,
                               benchmark,
                               time_stamp=timestamp(),
                               main_folder=self.main_folder)
      challenge.start()
      yield challenge

      # 日志更新
      mlogger.update()

      # 环境回收
      challenge.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Predict(self, exp_name, exp_param, **kwargs):
    # 1.step 配置基本信息
    self.__prepare_context(exp_name, exp_param, **kwargs)

    try:
      predict =\
        AntBatch(self,
                 exp_name,
                 self.api_token,
                 self.data_factory,
                 self.dump_dir,
                 self.task,
                 self.dataset)
      predict.start()
      yield predict

      # 日志更新
      mlogger.update()

      # 环境回收
      predict.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Ensemble(self, exp_name, exp_param, stage, **kwargs):
    # 1.step 配置基本信息
    self.__prepare_context(exp_name, exp_param, **kwargs)

    try:
      ensemble_handler = \
        AntEnsemble(self,
                    exp_name,
                    self.data_factory,
                    self.dump_dir,
                    self.api_token,
                    self.task,
                    self.dataset,
                    stage)
      # 准备ensemble控制
      ensemble_handler.start()
      yield ensemble_handler

      # 关闭recorder
      ensemble_handler.context.recorder.close()

      # 日志更新
      mlogger.update()

      # 环境回收
      ensemble_handler.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Demo(self, exp_name, exp_param, html_template=None,  **kwargs):
    # 1.step 配置基本信息
    self.__prepare_context(exp_name, exp_param, **kwargs)

    try:
      demo = AntDemo(self,
                     exp_name,
                     self.dump_dir,
                     self.api_token,
                     self.task,
                     html_template=html_template,
                     time_stamp=timestamp())
      # demo
      demo.start()
      yield demo

      # 日志更新
      mlogger.update()

      # 环境回收
      demo.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Browser(self, exp_name, exp_param, ip=None, port=10000, **kwargs):
    # 1.step 配置基本信息
    self.__prepare_context(exp_name, exp_param, **kwargs)

    try:
      browser = AntBrowser(self,
                           exp_name,
                           self.api_token,
                           self.data_factory,
                           kwargs.get('dataset', ''),
                           self.dump_dir)
      # 浏览控制
      browser.start()
      yield browser

      # 日志更新
      mlogger.update()

      # 环境回收
      browser.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Activelearning(self):
    pass


